'''
    Impl = Disj | Disj -> Impl
    Disj = Conj | Conj '|' Disj
    Conj = Neg  | Neg & Conj
    Neg = Lit | ! Lit
    Lit = [A-Z]  | ( Impl )
'''

def lexer(s):
    for c in s:
        if c==" ":
            continue
        yield c
    while True:
        yield '\0'

def pr(node):
    a = "("
    if node.left!=None:
        a += pr(node.left)
    a += " " + node.name + " "
    if node.right != None:
        a += pr(node.right)
    a+=')'
    return a

class Node:
    def __init__(self, left, right, name):
        self.left=left
        self.right=right
        self.name=name

class parser:

    def __init__(self, s):
        self.lex = lexer(s)
        self.current = next(self.lex)

    def parse(self):
        return self.Impl()

    def accept(self, c):
        if self.current == c:
            self.current = next(self.lex)
            return True
        return False

    def expect(self, c):
        if self.current == c:
            self.current = next(self.lex)
            return True
        print "Unexpected character", self.current, "expected", c
        return False

    def Impl(self):
        l = self.Disj()
        if self.accept('-'):
            if self.expect('>'):
                r = self.Impl()
                if r != None:
                    return Node(l, r, "implies")
            return None
        return l

    def Disj(self):
        l = self.Conj()
        if self.accept('|'):
            r = self.Disj()
            if r == None:
                return None
            return Node(l, r, "or")
        return l

    def Conj(self):
        l = self.Neg()
        if self.accept('&'):
            r = self.Conj()
            if r == None:
                return None
            return Node(l, r, "and")
        return l

    def Neg(self):
        if self.accept('!'):
            l = self.Lit()
            if l == None:
                return None
            return Node(None, l, "not")
        return self.Lit()

    def Lit(self):
        if self.accept('('):
            r = self.Impl()
            if self.expect(')'):
                return r
            return None
        l = self.current
        self.current = next(self.lex)
        if ord(l) < ord('A') or ord(l) > ord('Z'):
            print "Expected a capital letter"
            return None
        return Node(None, None, l)

def parse(s):
    p = parser(s)
    return p.parse()

if __name__ == "__main__":
    p = parser(raw_input("Enter a boolean formula: "))
    tree = p.Impl()
    if tree != None:
        print pr(tree)
    else:
        print "Failed to parse"
